#ifndef Image_h
#define Image_h

#include "config.h"

// Single pixel in an image
struct Pixel {
	unsigned char p[4];
};

struct DoublePixel {
	double p[3];

	DoublePixel() {}
	DoublePixel(double r, double g, double b)
	{
		p[0] = r;
		p[1] = g;
		p[2] = b;
	}

	DoublePixel(Pixel c)
	{
		p[0] = c.p[0];
		p[1] = c.p[1];
		p[2] = c.p[2];
	}

	DoublePixel operator* (double d)
	{
		return DoublePixel(p[0]*d, p[1]*d, p[2]*d);
	}

	DoublePixel& operator+= (DoublePixel c)
	{
		p[0] += c.p[0];
		p[1] += c.p[1];
		p[2] += c.p[2];
		return *this;
	}

	static inline unsigned char clamp(double d)
	{
		int i = (int)d;
		return i < 256 ? i : 255;
	}

	Pixel toPixel()
	{
		Pixel newP = {clamp(p[0]), clamp(p[1]), clamp(p[2]), 255};
		return newP;
	}
};

// Simple image class
class SimpleImage
{
public:
	SimpleImage() : w_(0), h_(0), data_(0) {}

	SimpleImage(int width, int height, Pixel* data = 0)
		: w_(width), h_(height)
	{
		if (data)
			data_ = data;
		else
			data_ = new Pixel[w_*h_];
	}

	~SimpleImage()
	{
		delete data_;
	}

	void snapshot(HDC hdc, int x, int y, int w, int h);
	void blit(SimpleImage* im);

	void resample(const SimpleImage& i);
	void resampleBilinear(const SimpleImage& i);

	// Return pixel at location x,y
	Pixel get(int x, int y) const
	{
		return pix(x,y);
	}

	Pixel getBilinear(double dX, double dY) const
	{
		int iX = (int)dX;
		int iY = (int)dY;

		DoublePixel p[2][2];
		p[0][0] = getBounded(iX,iY);
		p[1][0] = getBounded(iX+1,iY);
		p[0][1] = getBounded(iX,iY+1);
		p[1][1] = getBounded(iX+1,iY+1);

		double tX = dX - iX;  // Large when close to p[1][]
		double tY = dY - iY;  // Large when close to p[][1]

		DoublePixel totSum(0,0,0);
		totSum += p[0][0]*(1-tX)*(1-tY);
		totSum += p[1][0]*(  tX)*(1-tY);
		totSum += p[0][1]*(1-tX)*(  tY);
		totSum += p[1][1]*(  tX)*(  tY);

		return totSum.toPixel();
	}

	// Return pixel at location x,y bounded within image
	Pixel getBounded(int x, int y) const
	{
		/*if (x < 0) x = 0;
		if (y < 0) y = 0;
		if (x >= w_) x = w_-1;
		if (y >= h_) y = h_-1;*/
		//const Pixel black = {0, 0, 0, 255};
		// const Pixel back = {255,255,255,255};
		const Pixel back = {config.backRed, config.backGreen, config.backBlue, 255};
		if (x >= 0 && y >= 0 && x < w_ && y < h_)
			return get(x,y);
		else
			return back;
	}

	// Set pixel at location x,y to p
	void set(int x, int y, Pixel p)
	{
		pix(x,y) = p;
	}

	void clear(Pixel p);

	// Return width and height
	int w() const { return w_; }
	int h() const { return h_; }

	// Get raw pixel data
	Pixel* data() { return data_; }
	const Pixel* data() const { return data_; }

private:
	// Disallow copy constructor
	SimpleImage(const SimpleImage& i);

	const Pixel& pix(int x, int y) const
	{
		return data_[y*w_+x];
	}

	Pixel& pix(int x, int y)
	{
		return data_[y*w_+x];
	}

	int w_;
	int h_;
	Pixel* data_;
};

#endif

