#include <stdafx.h>

#include "Image.h"
#include "PrefetchTexture.h"
#include "TexturePool.h"

#include <assert.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <algorithm>
#include <memory>

#include "iSlideshow.h"

// Create large texture
PrefetchTexture::PrefetchTexture(SimpleImage* im, TexturePool* p)
{
	texid = 0;
	pool = p;

	// Resize image if necessary
	if (im->w() >= p->w() || im->h() >= p->w())
	{
		const int MAX_RESAMPLE = 1024;
		int maxTex = min(p->w(), MAX_RESAMPLE);
		maxTex = min(p->h(), maxTex);

		int wPix, hPix;
		double ratio = (double)im->w() / im->h();
		if (im->w() > im->h()) {
			wPix = min(im->w(),maxTex);
			hPix = wPix/ratio;
		} else {
			hPix = min(im->h(),maxTex);
			wPix = hPix*ratio;
		}

		SimpleImage* nim = new SimpleImage(wPix, hPix);
		/*int res = gluScaleImage(GL_RGBA,
			im->w(),  im->h(), GL_UNSIGNED_BYTE,  im->data(),
			nim->w(), nim->h(), GL_UNSIGNED_BYTE, nim->data());*/
		nim->resampleBilinear(*im);

		w = nim->w();
		h = nim->h();
		delete im;
		im = nim;
	}

	w = im->w();
	h = im->h();

	// Create border around image if necessary
	if (w < p->w() || h < p->h())
	{
		int w2 = (w < p->w()) ? (w + 1) : w;
		int h2 = (h < p->h()) ? (h + 1) : h;
		SimpleImage* nim = new SimpleImage(w2, h2);
		Pixel back = {config.backRed, config.backGreen, config.backBlue,0};
		nim->clear(back);
		nim->blit(im);
		delete im;
		im = nim;
	}

/*

	SimpleImage* nim = new SimpleImage(2048, 1600);
	nim->resample(*im);
	delete im;

	w = nim->w();
	h = nim->h();
*/
	//CreateMipmap(im);
	CreateMipmap(im);
}

PrefetchTexture::~PrefetchTexture()
{
	// Release textures
	if (texid)
		pool->release(texid);

	// Free images
	for (int i = 0; i < images.size(); i++)
	{
		delete images[i];
	}
}

// Downsample src to dst size
static void DownSample(SimpleImage& dst, const SimpleImage& src)
{
	// Determine multiple of size along x and y axis
	int xMult = (int)ceil((double)src.w()/dst.w());
	int yMult = (int)ceil((double)src.h()/dst.h());

	xMult = min(xMult, 2);
	yMult = min(yMult, 2);

	for (int y = 0; y < dst.h(); y++)
	{
		for (int x = 0; x < dst.w(); x++)
		{
			// Sum pixels
			int iPix[4] = {0, 0, 0, 0};
			for (int y2 = 0; y2 < yMult; y2++) {
				for (int x2 = 0; x2 < xMult; x2++)
				{
					Pixel srcPix = src.getBounded(x*xMult + x2, y*yMult + y2);
					for (int i = 0; i < 4; i++)
						iPix[i] += srcPix.p[i];
				}
			}

			// Divide out number of pixels
			Pixel pix;
			for (int i = 0; i < 4; i++)
				pix.p[i] = iPix[i] / (xMult*yMult);

			pix.p[3] = 255;
			dst.set(x,y,pix);
		}
	}
}

inline int closestPower(double x, double p)
{
	const double THRESHOLD = 0.001;
	double logX = log(x)/log(p);
	if (logX - fabs(logX) > THRESHOLD)
		return 1 << (int)fabs(logX);
	else
		return 1 << (int)ceil(logX);
	// return 1 << (int)ceil(logX);
}

void PrefetchTexture::CreateMipmap(SimpleImage* im)
{
	// Scale image for mipmaps
	if (pool->useMipmap())
	{
		// Construct data array
		images.resize(pool->numLevels());

		// Add border if room
		int borderX = (im->w() < pool->w()) ? 1 : 0;
		int borderY = (im->h() < pool->h()) ? 1 : 0;

		if (borderX + borderY > 0)
		{
			const Pixel back = {255,255,255,255};
			images[0] = new SimpleImage(im->w()+borderX, im->h()+borderY);
			int y;
			for (y = 0; y < im->h(); y++) {
				memcpy(images[0]->data() + images[0]->w()*y, im->data() + im->w()*y,
				       sizeof(Pixel)*im->w());

				if (borderX)
					images[0]->set(im->w(),y,back);
			}
			if (borderY) {
				memset(images[0]->data() + images[0]->w()*y, *(int*)&back,
					sizeof(Pixel)*images[0]->w());
			}

			delete im;
		} else {
			images[0] = im;
		}

		// Now, genereate mip map levels
		for (int level = 1; level < images.size(); ++level)
		{
			SimpleImage*& tiCur = images[level];
			SimpleImage*& tiLast = images[level-1];

			// Construct this level by shrinking the last level
			int curW = (int)ceil(tiLast->w() / 2.0);
			int curH = (int)ceil(tiLast->h() / 2.0);
			if (curW < 1) curW = 1;
			if (curH < 1) curH = 1;
			
			tiCur = new SimpleImage(curW, curH);

			DownSample(*tiCur, *tiLast);
		}
	} else {
		images.resize(1);
		images[0] = im;
	}
}

void PrefetchTexture::Draw()
{
	// Load image
	Load();

	glEnable(GL_TEXTURE_2D);

	float xB, yB, xE, yE;
	
	int screenDim[4];
	glGetIntegerv(GL_VIEWPORT, screenDim);
	float screenRatio = (float)screenDim[2]/screenDim[3];
	float imageRatio = (float)w / (float)h;
	float ratioRatio = imageRatio/screenRatio;

	float tW, tH;
	tW = ((float)w / pool->w());
	tH = ((float)h / pool->h());

	/*
	float MIN_THRESHOLD = 0.8;
	float MAX_THRESHOLD = 1.0/MIN_THRESHOLD;

	if (ratioRatio > MIN_THRESHOLD && ratioRatio < MAX_THRESHOLD)
	{
		// Use entire image if image ratio is close to screen ratio
		xB = yB = -1.0;
		xE = yE = 1.0;
	}*/
		
	if (screenRatio > imageRatio) {
		// Screen wider than image
		float b = 1.0 - ratioRatio;
		xB = -1.0 + b;
		xE = 1.0 - b;
		yB = -1.0;
		yE = 1.0;
	} else {
		// Screen taller than image
		float b = 1.0 - screenRatio/imageRatio;
		yB = -1.0 + b;
		yE = 1.0 - b;
		xB = -1.0;
		xE = 1.0;
	}

	glBindTexture(GL_TEXTURE_2D, texid);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0,tH);
		glVertex3f(xB,yE,0.0);

		glTexCoord2f(0.0,0.0);
		glVertex3f(xB,yB,0.0);

		glTexCoord2f(tW,0.0);
		glVertex3f(xE,yB,0.0);

		glTexCoord2f(tW,tH);
		glVertex3f(xE,yE,0.0);
	glEnd();
}


void PrefetchTexture::Load()
{
	if (texid)
		return;

	// Allocate texture
	texid = pool->obtain();
	glBindTexture(GL_TEXTURE_2D, texid);

	// Create texture
	glBindTexture(GL_TEXTURE_2D, texid);
	glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	for (int i = 0; i < images.size(); i++)
	{
		glPixelStorei(GL_UNPACK_ROW_LENGTH, images[i]->w());
		glTexSubImage2D(GL_TEXTURE_2D, i, 0, 0, 
			images[i]->w(), images[i]->h(), GL_RGBA, 
			GL_UNSIGNED_BYTE, images[i]->data());

		delete images[i];
		images[i] = 0;
	}

	glPopClientAttrib();
}
